use anyhow::{anyhow, Result};
use reqwest::header::{HeaderMap, HeaderName, HeaderValue};
use sha3::Digest;
use std::collections::HashMap;
use std::io::{self, BufReader, BufWriter};
use std::str::FromStr;
use std::{
    env,
    ffi::OsStr,
    path::{Path, PathBuf},
};
use tokio::io::{AsyncReadExt, AsyncWriteExt};

pub mod model;

async fn build_request(
    source: &model::Source,
    srcpath: Option<&Path>,
) -> Result<(reqwest::Client, reqwest::Request)> {
    let mut request_builder =
        reqwest::Client::new().request(source.method.0.clone(), source.url.clone());
    let headers: Result<Vec<(_, _)>> = source
        .headers
        .clone()
        .into_iter()
        .map(|(k, v)| {
            let hn = HeaderName::from_str(&k)?;
            let hv = HeaderValue::from_str(&v)?;
            Ok((hn, hv))
        })
        .collect();
    request_builder = request_builder.headers(HeaderMap::from_iter(headers?));
    request_builder = request_builder.query(&source.query);
    request_builder = match &source.auth {
        None => request_builder,
        Some(auth) => match auth {
            model::Auth::Basic(b) => {
                request_builder.basic_auth(b.username.clone(), Some(b.password.clone()))
            }
            model::Auth::Bearer(b) => request_builder.bearer_auth(b),
        },
    };
    if let Some(source_body) = &source.body {
        if let Some(srcpath_) = srcpath {
            request_builder = match source_body {
                model::Body::Raw(rb) => {
                    let mut buffer = Vec::<u8>::new();
                    let mut file = tokio::fs::File::open(srcpath_.to_owned().join(rb)).await?;
                    let read = file.read_buf(&mut buffer).await?;
                    buffer.truncate(read);
                    request_builder.body(buffer)
                }
                model::Body::Form(fb) => {
                    let mut buffer = Vec::<u8>::new();
                    let mut file = tokio::fs::File::open(srcpath_.to_owned().join(fb)).await?;
                    let read = file.read_buf(&mut buffer).await?;
                    buffer.truncate(read);
                    request_builder
                        .form(&serde_json::from_slice::<HashMap<String, String>>(&buffer)?)
                }
                model::Body::Multipart(fbvec) => {
                    let mut form = reqwest::multipart::Form::new();
                    for mb in fbvec {
                        let mb_ = srcpath_.to_owned().join(mb);
                        let mut buffer = Vec::<u8>::new();
                        let mut file = tokio::fs::File::open(&mb_).await?;
                        let read = file.read_buf(&mut buffer).await?;
                        buffer.truncate(read);
                        let name = mb_
                            .file_name()
                            .ok_or_else(|| anyhow!("Failed to get file name"))?
                            .to_str()
                            .ok_or_else(|| anyhow!("Failed to get file name"))?
                            .to_string();
                        form = form.part(name, reqwest::multipart::Part::bytes(buffer));
                    }
                    request_builder.multipart(form)
                }
            };
        }
    }
    let (client, request) = request_builder.build_split();
    let request: reqwest::Request = request?;
    Ok((client, request))
}

async fn check(input: model::CheckInput) -> Result<model::CheckOutput> {
    let (client, request) = build_request(&input.source, None).await?;
    let mut response = client.execute(request).await?;
    Ok(model::CheckOutput(
        match input.source.version_check_method {
            model::VersionCheckMethod::None => vec![model::Version::None],
            model::VersionCheckMethod::Hash => {
                let mut hasher = sha3::Sha3_256::new();
                while let Some(chunk) = response.chunk().await? {
                    hasher.update(&chunk);
                }
                let hash = hasher.finalize();
                vec![model::Version::Hash(hex::encode(hash))]
            }
            model::VersionCheckMethod::Header(headername) => {
                let headervalue: &HeaderValue = response
                    .headers()
                    .get(headername)
                    .ok_or_else(|| anyhow!(""))?;
                vec![model::Version::Header(headervalue.to_str()?.to_string())]
            }
        },
    ))
}

async fn in_(input: model::InInput, dest: &Path) -> Result<model::InOutput> {
    let input: model::InInput = model::InInput {
        source: input.source.merge_params(&input.params),
        version: input.version,
        params: input.params,
    };
    let (client, request) = build_request(&input.source, None).await?;
    let mut response = client.execute(request).await?;

    let dest_code = dest.to_owned().join("code");
    let dest_headers = dest.to_owned().join("headers");
    let dest_bytes = dest.to_owned().join("output");

    let mut dest_code_writer = tokio::io::BufWriter::new(tokio::fs::File::create(dest_code).await?);
    dest_code_writer
        .write_all(response.status().to_string().as_bytes())
        .await?;
    dest_code_writer.flush().await?;

    let response_headers: Result<Vec<(String, String)>> = response
        .headers()
        .iter()
        .map(|(headername, headervalue)| {
            Ok((
                headername.as_str().to_string(),
                headervalue.to_str()?.to_string(),
            ))
        })
        .collect();
    let mut dest_headers_writer =
        tokio::io::BufWriter::new(tokio::fs::File::create(dest_headers).await?);
    dest_headers_writer
        .write_all(serde_json::to_vec(&response_headers?)?.as_ref())
        .await?;

    let mut hasher = sha3::Sha3_256::new();
    let mut dest_bytes_writer =
        tokio::io::BufWriter::new(tokio::fs::File::create(dest_bytes).await?);
    while let Some(chunk) = response.chunk().await? {
        if input.source.version_check_method == model::VersionCheckMethod::Hash {
            hasher.update(&chunk);
        }
        dest_bytes_writer.write_all(&chunk).await?;
    }
    let hash = hasher.finalize();
    dest_bytes_writer.flush().await?;

    let out = model::InOutput {
        version: match input.source.version_check_method {
            model::VersionCheckMethod::None => model::Version::None,
            model::VersionCheckMethod::Hash => model::Version::Hash(hex::encode(hash)),
            model::VersionCheckMethod::Header(headername) => {
                let headervalue: &HeaderValue = response
                    .headers()
                    .get(headername)
                    .ok_or_else(|| anyhow!(""))?;
                model::Version::Header(headervalue.to_str()?.to_string())
            }
        },
        metadata: vec![],
    };
    if out.version != input.version {
        return Err(anyhow!(format!(
            "Did not get correct version of resource, requested '{:?}', got '{:?}'",
            input.version, out.version
        )));
    }
    Ok(out)
}

async fn out(input: model::OutInput, src: &Path) -> Result<model::OutOutput> {
    let input = model::OutInput {
        source: input.source.merge_params(&input.params),
        params: input.params,
    };
    let (client, request) = build_request(&input.source, Some(src)).await?;
    let response = client.execute(request).await?;
    response.error_for_status()?;
    let version = check(model::CheckInput {
        source: input.source.clone(),
        version: model::Version::None,
    })
    .await?
    .0[0]
        .clone();
    Ok(model::OutOutput {
        version,
        metadata: vec![],
    })
}

#[tokio::main]
async fn main() -> Result<()> {
    let mut args = env::args();
    let operation: PathBuf = args
        .next()
        .ok_or_else(|| anyhow!("No binary name! This should never happen"))?
        .into();
    let operation: &OsStr = operation
        .file_name()
        .ok_or_else(|| anyhow!("Could not get binary name! This should never happen"))?;
    if operation == "check" {
        let input: model::CheckInput = serde_json::from_reader(BufReader::new(io::stdin()))?;
        let r = check(input).await?;
        serde_json::to_writer(BufWriter::new(io::stdout()), &r)?;
        Ok(())
    } else if operation == "in" {
        let input: model::InInput = serde_json::from_reader(BufReader::new(io::stdin()))?;
        let path: PathBuf = args
            .next()
            .ok_or_else(|| anyhow!("No destination specified!"))?
            .into();
        let r = in_(input, &path).await?;
        serde_json::to_writer(BufWriter::new(io::stdout()), &r)?;
        Ok(())
    } else if operation == "out" {
        let input: model::OutInput = serde_json::from_reader(BufReader::new(io::stdin()))?;
        let path: PathBuf = args
            .next()
            .ok_or_else(|| anyhow!("No destination specified!"))?
            .into();
        let r = out(input, &path).await?;
        serde_json::to_writer(BufWriter::new(io::stdout()), &r)?;
        Ok(())
    } else {
        Err(anyhow!("Invalid binary name!"))
    }
}
