use anyhow::Result;
use serde::{Deserialize, Deserializer, Serialize, Serializer};
use std::path::PathBuf;
use std::str::FromStr;

#[derive(Clone, Debug, Default, PartialEq, Eq)]
pub struct Method(pub reqwest::Method);

impl From<reqwest::Method> for Method {
    fn from(item: reqwest::Method) -> Self {
        Method(item)
    }
}

impl<'de> Deserialize<'de> for Method {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        #[derive(Deserialize)]
        struct MethodData(String);

        let data = MethodData::deserialize(deserializer)?;
        let r_m = reqwest::Method::from_str(&data.0.to_uppercase());
        match r_m {
            Ok(r_m) => Ok(r_m.into()),
            Err(_e) => Err(serde::de::Error::custom(format!(
                "No such method '{}'",
                data.0
            ))),
        }
    }
}

impl Serialize for Method {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        serializer.serialize_str(&self.0.to_string())
    }
}

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq)]
pub struct AuthBasic {
    pub username: String,
    pub password: String,
}

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq)]
#[serde(rename_all = "snake_case")]
pub enum Auth {
    Basic(AuthBasic),
    Bearer(String),
}

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq)]
#[serde(rename_all = "snake_case")]
pub enum Body {
    Raw(PathBuf),
    Form(PathBuf),
    Multipart(Vec<PathBuf>),
}

#[derive(Serialize, Deserialize, Clone, Debug, Default, PartialEq, Eq)]
#[serde(rename_all = "snake_case")]
pub enum VersionCheckMethod {
    Header(String),
    #[default]
    Hash,
    None,
}

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq)]
pub struct Source {
    pub url: url::Url,
    pub method: Method,
    #[serde(default)]
    pub headers: Vec<(String, String)>,
    #[serde(default)]
    pub query: Vec<(String, String)>,
    #[serde(default)]
    pub auth: Option<Auth>,
    #[serde(default)]
    pub body: Option<Body>,
    #[serde(default)]
    pub version_check_method: VersionCheckMethod,
}

impl Source {
    pub fn merge_params(self, params: &Params) -> Self {
        let url = self.url.clone();
        let mut method: Method = self.method.clone();
        let mut headers: Vec<(String, String)> = self.headers.clone();
        let mut query: Vec<(String, String)> = self.query.clone();
        let mut auth: Option<Auth> = self.auth.clone();
        let mut body: Option<Body> = self.body.clone();
        let version_check_method = self.version_check_method.clone();
        if let Some(params_method) = &params.method {
            method = params_method.clone();
        }
        if let Some(params_headers) = &params.headers {
            headers.extend(params_headers.clone());
        }
        if let Some(params_query) = &params.query {
            query.extend(params_query.clone());
        }
        if let Some(params_auth) = &params.auth {
            auth = Some(params_auth.clone());
        }
        if let Some(params_body) = &params.body {
            body = Some(params_body.clone());
        }

        Self {
            url,
            method,
            headers,
            query,
            auth,
            body,
            version_check_method,
        }
    }
}

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq)]
#[serde(rename_all = "snake_case")]
pub enum Version {
    Header(String),
    Hash(String),
    None,
}

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq)]
pub struct Metadata {
    pub name: String,
    pub value: String,
}

#[derive(Serialize, Deserialize, Clone, Debug, Default, PartialEq, Eq)]
pub struct Params {
    pub method: Option<Method>,
    pub headers: Option<Vec<(String, String)>>,
    pub query: Option<Vec<(String, String)>>,
    pub auth: Option<Auth>,
    pub body: Option<Body>,
}

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq)]
pub struct CheckInput {
    pub source: Source,
    pub version: Version,
}

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq)]
pub struct InInput {
    pub source: Source,
    pub version: Version,
    pub params: Params,
}

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq)]
pub struct OutInput {
    pub source: Source,
    pub params: Params,
}

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq)]
pub struct CheckOutput(pub Vec<Version>);

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq)]
pub struct InOutput {
    pub version: Version,
    pub metadata: Vec<Metadata>,
}

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq)]
pub struct OutOutput {
    pub version: Version,
    pub metadata: Vec<Metadata>,
}
